"""
	Functions for parsing sgf files.
"""

__author__ = "Roland Lezuo <roland.lezuo@chello.at>"
__licence__ = "GNU General Public License Version 2"


import logging
logger = logging.getLogger("sgfparser")

from pyparsing import *

#	*****************************************************************************
#	This part of the parser handles the low level stuff and returs a _Collection
#	which nobody wants to see upstairs, second part of the parser handles adds
#	all the chrome for the higher level functions
#	*****************************************************************************


class _Node(dict):
	"""
	Helper class, used to catch together Nodes with more than one property
	"""
	def __init__(self):
		self._next = []
		self._prev = None
	def __str__(self):
		if len(self._next) == 1:
			return "%s -> %s" % (self.keys(), self._next)
		else:
			s = ""
			for n in self._next:
				r = str(n)
				s = s + r 
			return "%s -> %s" % (self.keys(), s)
class _Collection:
	"""
	Helper class, returned by the parser
	"""
	def __init__(self):
		self._games = []
	def get_games(self):
		"""
		Returns a list of all games contained in the collection
		"""
		return self._games
	def __str__(self):
		return "_Collection with %s games" % len(self._games)

				

def action_PropValue(string, loc, tokens):
	"""	
	when parsing a property value we have to complicated handle escaped closeing brackts (\]).
	therefore the string gets sliced, remove this slicing here.
	"""
	s = ""
	for t in tokens:
		 s = s + t
	return [s]

def action_LiteralEscBrack(string, loc, tokens):
	"""
	when matching an escaped closing bracket (\]) remove the escaping character
	"""
	return [']']

def action_Property(string, loc, tokens):
	if (len(tokens) == 1):
		# token with no content
		return [ (tokens[0], []) ]
	return [ (tokens[0], tokens[1:]) ]

def action_Node(string, loc, tokens):
	"""
	creates a dictionary with all properties of the node
	node[PropIdent] = PropValue
	"""
	node = _Node()
	for t in tokens:
		assert(len(t) == 2)
		node[t[0]] = t[1]
	return node

def action_Sequence(string, loc, tokens):
	"""
	makes a double linked list of the _Nodes
	"""
	old = None
	root = None
	for t in  tokens:
		if root == None: root = t
		if old != None: old._next.append(t)
		t._prev = old
		old = t
	return root

def action_GameTree(string, loc, tokens):
	root = None
	for t in  tokens:
		if root == None: root = t
		else: root._next.append(t)
	return root


def action_Collection(string, loc, tokens):
	col = _Collection()
	for t in  tokens:
		col._games.append(t)
	return col



# pyparsing declaration of the SGF dataformat
# according to http://www.red-bean.com/sgf/sgf4.html#2
# EBNF

#PropValue: the value of a property, flattened because of the ugly \] handling 
PropValue  = OneOrMore( Or( CharsNotIn( "\\]" ) ^ Literal("\\]").setParseAction(action_LiteralEscBrack) ^ Literal("\\") ) ).setParseAction(action_PropValue)
PropIdent  = Word( "ABCDEFGHIJKLMNOPQRSTUVWXYZ" )

#Property: Property Identifier + its value, remove [] around value, we don't need it
Property   = (PropIdent + OneOrMore( Literal("[").suppress() + Optional(PropValue) + Literal("]").suppress() ) ).setParseAction(action_Property)

#EmtpyNode: actually forbidden but many files contain them neitherless, suppress them silently.
EmptyNode  = Literal(";").suppress()
Node       = ( Literal(";").suppress() + OneOrMore( Property ) ).setParseAction(action_Node)
Sequence   = OneOrMore( Node | EmptyNode ).setParseAction(action_Sequence)
GameTree   = Forward()
GameTree  << ( Literal("(").suppress() + Sequence + ZeroOrMore( GameTree ) + Literal(")").suppress() ).setParseAction(action_GameTree)
Collection = OneOrMore( GameTree ).setParseAction(action_Collection)



#	*****************************************************************************
#	High level chrome
#	*****************************************************************************

class GnomeGoGame:
	"""
	This class is the highest level chrome provided, it takes a SGFGame as input and provides
	functions to step forward and backward, always updating the current board position.
	Currently it only provides this for the mainline of a game.
	"""
	def __init__(self, g):
		logger.debug("Creating GnomeGoGame")
		self._game = g
		self._mh = {}		# move history, dictionary, keys = board_position value = movenumber
		self._bs = set([])	# black stones in the actual state
		self._ws = set([])	# white stones in the actual state
		self._mn = 1		# next move to happen
		self._maxmn = 1 + len(self._game.get_mainline())	# number of moves of the whole game

	def backstep(self):
		raise "TODO"
	
	def step(self):
		if self._mn == self._maxmn: 
			return False

		# get next move from sequence of moves and put stone 
		node = self._game.get_mainline()[self._mn-1]

		stone = node._coord
		if node._color == 'B':
			self._bs |= set([stone])
		elif node._color == 'W':
			self._ws |= set([stone])
		else:
			assert("Unknown color in mainline")

		# update move history dictionary, this can never be wrong
		if self._mh.has_key(stone):
			self._mh[stone].append(self._mn)
		else:
			self._mh[stone] = [self._mn]
	
		# put the last move into the history as well
		self._mh["last"] = stone

		# identify dead groups and remove these stones from actual state
		self._deadstones(node)
		self._mn += 1
		return True

	def get_white_stones(self):
		return self._ws

	def get_black_stones(self):
		return self._bs

	def get_move_history(self):
		return self._mh
	
	def get_metadata(self):
		return self._game._metadata

	# plays the new stone and removes all stones captured by this move
	# XXX hardcoded japanese rules
	def _deadstones(self, new):
		# get a list of all fields possibly affected by the new move
		neigbors = self._get_neigbors(new._coord)

		# look at these fields an look if an enemy dragon is there
		for n in neigbors:
			if new._color == 'B':
				# this was a black stone, check state of white groups
				if n in self._ws:
					dragon, libs = self._get_dragon_and_liberties(n, self._ws, self._bs)
					if libs == 0: 
						self._ws = self._ws.difference(dragon)
			else:
				if n in self._bs:
					dragon, libs = self._get_dragon_and_liberties(n, self._bs, self._ws)
					if libs == 0: 
						self._bs = self._bs.difference(dragon)



	# iterates the whole dragon the given stone is part of and counts liberties
	def _get_dragon_and_liberties (self, stone, friends, foes):
		group, liberties = self._get_dragon_and_liberties_rec(stone, set([]), set([]), friends, foes)
		#print "Dragon " + str(group) + " libs: " + str(liberties)
		return group, len(liberties)

	def _get_dragon_and_liberties_rec(self, stone, group, liberties, friends, foes):
		# find all neighbors
		group |= set([stone])
		neigbors = self._get_neigbors(stone)
		for n in neigbors:
			if n in group:
				pass				# alreay visited
			elif n in friends:
				g, l = self._get_dragon_and_liberties_rec(n, group, liberties, friends, foes)
				group |= g
				liberties |= l
			elif n in foes:
				pass				# does not add a liberty :)
			else:
				liberties |= set([n])		# add found field ot set of liberties
		return group, liberties
				
	def _get_neigbors(self, s):
		x,y = s
		ret = []	# list of field to be returned
		if x != 0:				ret.append((x-1, y))	#left
		if x != self._game._metadata.size-1:	ret.append((x+1, y))	#right
		if y != 0:				ret.append((x, y-1))	#top
		if y != self._game._metadata.size-1:	ret.append((x, y+1))	#bottom
		return ret
		
class SGFGameMeta:
	"""
	Capsulating all metadata of a game known by the parser.
	"""
	size = 19	# some default value
	player_white = 'unknown'
	rank_white = 'unkown'
	player_black = 'unknown'
	rank_black = 'unknown'
	result = 'unknown'
	event = 'unknown'
	date = 'unknown'	#YYYY-MM-DD
	round = 'unknown'
	rules = 'unknown'
	place = 'unknown'
	timelimit = 'unknown'

	def __str__(self):
		return 'W: ' + self.player_white + '(' + self.rank_white + ')' + ' vs B: ' + self.player_black + '(' + self.rank_black + ') ' +\
					self.date + '@' + self.event + ' ' + self.result

class SGFMoveNode:
	"""
	A move node is an abstraction of a stone placed on the board. 
	"""
	def __init__(self):
		self._next = []
		self._prev = None
		self._color = None
		self._number = None
		self._comment = None
		self._coord = None
	def add(self, var):
		self._next.append(var)
	def __str__(self):
		if len(self._next) == 0:
			return "%s%s" % (self._color, self._coord)
		elif len(self._next) == 1:
			return "%s%s, %s" % (self._color, self._coord, str(self._next[0]))
		else:
			s = ""
			c = 1
			for i in self._next:
				s = s + "\n[%s/%s...%s]" % (c, len(self._next), str(i))
				c = c + 1
			
			return "%s%s !!! %s" % (self._color, self._coord, s)

class SGFGame:
	"""
	This class is an abstraction of a whole game. Basically capsulates metadta and moves of a game.
	It also contains logic to create a mainline of a game, this is the game without possible variations.
	"""
	_moves = None
	_bm = None
	_wm = None
	_mainline = None
	_metadata = None

	def __init__(self):
		self._metadata = SGFGameMeta()

	# return the whole mainline
	def get_mainline(self):
		if self._mainline == None:
			self._create_mainlines()
		return self._mainline

	# return the main line
	def black_moves_mainline(self):
		if self._bm == None:
			self._create_mainlines()
		return self._bm

	# return the main line
	def white_moves_mainline(self):
		if self._wm == None: 
			self._create_mainlines()
		return self._wm

	def _create_mainlines(self):
		self._bm, self._wm, self._mainline = self._mainlining(self._moves)
		logger.debug("_create_mainlines white: " + str(self._wm))
		logger.debug("_create_mainlines black: " + str(self._bm))
	
	# wanders the SGFMoveNode double linked list along next[0] till it
	# finds an end
	# return:  a tuple
	#	list of white coords forming the white mainline
	#	list of black coords forming the black mainlibne	
	#	list of nodes forming the mainline
	def _mainlining(self, moves):
		wm = []
		bm = []
		mm = []
		node = moves
		logger.debug("_mainlining: %s", node)
		while len(node._next):
			if node._color == 'B':
				bm.append(node._coord)
			elif node._color == 'W':
				wm.append(node._coord)
			else:
				raise "Unknown move color %s" % node._color
			mm.append(node)
			node = node._next[0]
		else:
			# handle the last node too, warning, this could also be the empty list case
			if node:
				if node._color == 'B':
					bm.append(node._coord)
				elif node._color == 'W':
					wm.append(node._coord)
				else:
					raise "Unknown move color %s" % node._color
				mm.append(node)
		return (bm, wm, mm)
	
	def __str__(self):
		return  "SGFGame\n" \
		 	"   metadata: %s\n"  \
		 	"   moves: %s\n " % (str(self._metadata), str(self._moves))


class SGFParser:
	def __init__(self):
		self._collection = None
	def parseString(self, string):
		self._collection = Collection.parseString(string)
		assert(len(self._collection) == 1)
	def parseFile(self, file):
		self._collection = Collection.parseFile(file)
		assert(len(self._collection) == 1)
	def getNumberOfSGFGames(self):
		""" return the number of games in the collection parsed """
		if not self._collection:
			return 0
		else:
			return len(self._collection[0].get_games())
	def getSGFGame(self, number):
		""" return the number game in the collection parsed. Counting begins with 0. """
		if not self._collection:
			return None
		else:
			return create_SGFGame(self._collection[0], number)

def update_metadata(md, k, v):
	"""
	This functions adds a key k with a list of values v to the metadata is k is a known metadata property.
	Return true when the key was recognized, false else
	"""
	if   k == 'CA': pass
	elif k == 'GM':
		if v[0] != "1": raise "No GO record"
		return True
	elif k == 'FF':
		# version of the the sgf fiel
		pass
		return True
	elif k == 'SZ':
		md.size = int(v[0])
		return True
	elif k == 'PW':
		md.player_white = v[0]
		return True
	elif k == 'PB':
		md.player_black = v[0]
		return True
	elif k == 'WR':
		md.rank_white = v[0]
		return True
	elif k == 'BR':
		md.rank_black = v[0]
		return True
	elif k == 'KM':
		md.komi = v[0]
		return True
	elif k == 'RE':
		md.result = v[0]
		return True
	elif k == 'EV':
		md.event = v[0]
		return True
	elif k == 'DT':
		md.date = v[0]
		return True
	elif k == 'RO':
		md.round = v[0]
		return True
	elif k == 'RU':
		md.rules = v[0]
		return True
	elif k == 'PC':
		md.place = v[0]
		return True
	elif k == 'TM':
		md.timelimit = v[0]
		return True
	else:
		return False
			
def _move_to_logical(move, size):
	"""
	Converts SGF position encoding into a tuple of integers (x,y), with (-1,-1) beeing the PASS move
	"""
	m = move.upper()
	if len(m) == 0: return (-1, -1)
	if (size <= 19) and m == 'TT': return(-1, -1)
	x = ord(m[0]) - ord('A')
	y = ord(m[1]) - ord('A')

	return (x,size-y-1)

def create_SGFMoveNodes(node, oldnode, size, startnumber):
	"""
	creates a doulbe linked list of SGFMoveNodes from a doulbe linked _Node list
	it also checks if the properties are okay and rejects sgf setups not handled
	needs size of board to handle move conversion properly
	startnumber is the move number of the first move encountered
	FIXME: when encountering metadata they should be propergated into game metadata, therefore
	gamemetadata would need to be argument as well
	FIXME: almost all properties of the node are thrown away only C are kept
	"""
	assert(isinstance(node, _Node))
	if oldnode: assert(isinstance(oldnode, SGFMoveNode))

	rootnode = SGFMoveNode()
	rootnode._number = startnumber
	rootnode._prev = oldnode
	try:
		rootnode._comment = node['C']
	except:
		pass

	if node.has_key('B'):
		rootnode._color = 'B'
		rootnode._coord = _move_to_logical(node['B'][0], size)
	elif node.has_key('W'): 
		rootnode._color='W'
		rootnode._coord = _move_to_logical(node['W'][0], size)
	else:
		logger.warning("create_SGFMoveNodes: Node does not contain move during game play, pruning tree!")
		# XXX we skip this node
		oldnode._comment = "This is your parser speaking, I had troubles parsing the gametree and cut off here! Do not trust me!"
		return None
	
	for n in node._next:
		nodes = create_SGFMoveNodes(n, rootnode, size, startnumber+1)
		if nodes:
			nodes._prev = rootnode
			rootnode._next.append(nodes)

	return rootnode

def create_SGFMoveNodes_from_AB(game, setup):
	"""
	Input, a SGFGame and the Porperty value of the AB property, return: double linked
	list of SGFMoveNodes
	"""
	return _create_SGFMoveNodes_from_AX('B', setup, game._metadata.size)
def create_SGFMoveNodes_from_AW(game, setup):
	"""
	Input, a SGFGame and the Porperty value of the AW property, return: double linked
	list of SGFMoveNodes
	"""
	return _create_SGFMoveNodes_from_AX('W', setup, game._metadata.size)

def _create_SGFMoveNodes_from_AX(color, stones, size):
	"""
	Create doulbe linked list of SGFMoveNodes taking color, a list of SGF coordinates and an oldnode for chaining
	return head of list 
	"""

	rootnode = None
	oldnode = None

	for s in stones:
		if not rootnode:
			rootnode = SGFMoveNode()
			node = rootnode
		else:
			node = SGFMoveNode()
			if oldnode: oldnode._next.append(node)

		node._prev = oldnode

		node._number = 0
		node._color = color
		node._coord = _move_to_logical(s, size)
		oldnode = node

	return rootnode


def create_SGFGame(coll, number):
	"""
	creates a SGFGame from a _Collection and a given game number.
	"""
	game = SGFGame()
	g = coll.get_games()[number]

	assert(isinstance(g, _Node))

	ab = None
	aw = None
	for k,v in g.iteritems():
		# first check if the property is a metadata stuff, and merge it into games metadata if so
		wasmeta = update_metadata(game._metadata, k, v)
		if wasmeta: 
			logger.info("Consumed metadata %s" % k)
			continue

		# no, was no metadata, check if it are moves

		if   k == 'AB': ab = create_SGFMoveNodes_from_AB(game, v)
		elif k == 'AW': aw = create_SGFMoveNodes_from_AW(game, v)
		elif k == 'B': raise "Unhandeled: moves in setup rootnode"
		elif k == 'W': raise "Unhandeled: moves in setup rootnode"
		else:
			logger.warning("Unknown property %s" % k)

	# successfully handled the rootnode, check if we had setup moves
	# XXX this is only correct as long as we do not handle moves in setup node
	if ab:
		game._moves = ab
	if aw:
		if not game._moves:
			# that is easy
			game._moves = aw
		else:
			node = game._moves
			while node._next: node = node._next[0]	# find end of setup move list
			print node
			node._next.append(aw)				# and attach white setup moves

	# as we propably had setup moves, all other moves have to be appended after these moves
	newroot = None
	if game._moves:
		newroot = game._moves
		while newroot._next: 
			assert(len(newroot._next) == 1)
			newroot = newroot._next[0]


	# now handle all children, we FALSELY assume that there are only move nodes to follow
	for n in g._next:
		#each variation start as move number 1
		nodes = create_SGFMoveNodes(n, None, game._metadata.size, 1)

		if nodes:
			if newroot:
				newroot._next.append(nodes)
			else:
				if game._moves: game._moves._next.append(nodes)
				else: game._moves = nodes
	# that's all
	return game



def main():
	"""
		Main function, runs tests on sgf files and directories when called
	"""
	import sys
	import os
	import stat

	if len(sys.argv) != 2:
		print "Usage: python " + str(__file__) + " path/to/sgffiles"
		print "If arg is directory: recurse and parse all files found, print summary"
		print "If arg is file: parse that sepcific file, verbose"
		sys.exit(-1)

	if stat.S_ISDIR(os.stat(sys.argv[1])[stat.ST_MODE]):
		# walk directory and print summary
		failed = []
		for root,dirs,files in os.walk(sys.argv[1]):
			for f in files:
				file = os.path.join(root, f)
				result_string = "%s parsing..." % file
				try:
					result = Collection.parseFile(file)
				except:
					result_string += "FAILED"
					failed.append(file)
					print result_string
					continue
				result_string += "OK gnomegogamecreation (game 0 only)..."
				try:
					create_SGFGame(result[0], 0)
				except:
					result_string += "FAILED"
					failed.append(file)
					print result_string
					continue
				result_string += "OK"
				print result_string
					
	
		print "-----------------SUMMARY"
		print "Failed files:"
		print failed
	else:
		try:
			result = Collection.parseFile(sys.argv[1])

			print "PARSING RESULT:"
			print result[0]
			print create_SGFGame(result[0], 0)
		except ParseException, e:
			print "ParseException: %s" % e
		


if __name__ == "__main__":
	import os
	if os.getenv("GNOMEGO_DEBUG") != None:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig()
	main()
