__author__ = "Roland Lezuo <roland.lezuo@chello.at>"
__licence__ = "GNU General Public License Version 2"

# implementation of an sgf file parser
# i did not like the one http://gotools.sourceforge.net/

import logging
logger = logging.getLogger("sgfparser")

import simpleparse.parser
import simpleparse.processor
import simpleparse.dispatchprocessor

from simpleparse.common import strings


# this is a helper class taking a SGFGame as input and caluclates the game progress from the
# moves made by both players. This includes: removing dead stones
# it offers a step method which calculates the next "state" from the recorded "events"
class GnomeGoGame:
	def __init__(self, g):
		logger.debug("Creating GnomeGoGame")
		self._game = g[0][0]
		self._mh = {}		# move history, dictionary, keys = board_position value = movenumber
		self._bs = set([])	# black stones in the actual state
		self._ws = set([])	# white stones in the actual state
		self._mn = 1		# next move to happen
		self._maxmn = 1 + len(self._game.get_mainline())	# number of moves of the whole game

	def backstep(self):
		raise "TODO"
	
	def step(self):
		if self._mn == self._maxmn: 
			return False

		# get next move from sequence of moves and put stone 
		node = self._game.get_mainline()[self._mn-1]

		stone = node._coord
		if node._color == 'B':
			self._bs |= set([stone])
		elif node._color == 'W':
			self._ws |= set([stone])
		else:
			assert("Unknown color in mainline")

		# update move history dictionary, this can never be wrong
		if self._mh.has_key(stone):
			self._mh[stone].append(self._mn)
		else:
			self._mh[stone] = [self._mn]

		# identify dead groups and remove these stones from actual state
		self._deadstones(node)
		self._mn += 1
		return True

	def get_white_stones(self):
		return self._ws

	def get_black_stones(self):
		return self._bs

	def get_move_history(self):
		return self._mh
	
	def get_metadata(self):
		return self._game.metadata

	# plays the new stone and removes all stones captured by this move
	# XXX hardcoded japanese rules
	def _deadstones(self, new):
		# get a list of all fields possibly affected by the new move
		neigbors = self._get_neigbors(new._coord)

		# look at these fields an look if an enemy dragon is there
		for n in neigbors:
			if new._color == 'B':
				# this was a black stone, check state of white groups
				if n in self._ws:
					dragon, libs = self._get_dragon_and_liberties(n, self._ws, self._bs)
					if libs == 0: 
						self._ws = self._ws.difference(dragon)
			else:
				if n in self._bs:
					dragon, libs = self._get_dragon_and_liberties(n, self._bs, self._ws)
					if libs == 0: 
						self._bs = self._bs.difference(dragon)



	# iterates the whole dragon the given stone is part of and counts liberties
	def _get_dragon_and_liberties (self, stone, friends, foes):
		group, liberties = self._get_dragon_and_liberties_rec(stone, set([]), set([]), friends, foes)
		#print "Dragon " + str(group) + " libs: " + str(liberties)
		return group, len(liberties)

	def _get_dragon_and_liberties_rec(self, stone, group, liberties, friends, foes):
		# find all neighbors
		group |= set([stone])
		neigbors = self._get_neigbors(stone)
		for n in neigbors:
			if n in group:
				pass				# alreay visited
			elif n in friends:
				g, l = self._get_dragon_and_liberties_rec(n, group, liberties, friends, foes)
				group |= g
				liberties |= l
			elif n in foes:
				pass				# does not add a liberty :)
			else:
				liberties |= set([n])		# add found field ot set of liberties
		return group, liberties
				
	def _get_neigbors(self, s):
		x,y = s
		ret = []	# list of field to be returned
		if x != 0:				ret.append((x-1, y))	#left
		if x != self._game.metadata.size-1:	ret.append((x+1, y))	#right
		if y != 0:				ret.append((x, y-1))	#top
		if y != self._game.metadata.size-1:	ret.append((x, y+1))	#bottom
		return ret
			

	


# containts metadata about a Game
class SGFGameMeta:
	size = 19	# some default value
	player_white = 'unknown'
	rank_white = 'unkown'
	player_black = 'unknown'
	rank_black = 'unknown'
	result = 'unknown'
	event = 'unknown'
	date = 'unknown'	#YYYY-MM-DD
	round = 'unknown'
	rules = 'unknown'
	place = 'unknown'
	timelimit = 'unknown'

	def __str__(self):
		return 'W: ' + self.player_white + '(' + self.rank_white + ')' + ' vs B: ' + self.player_black + '(' + self.rank_black + ') ' +\
					self.date + '@' + self.event + ' ' + self.result

# a choicepoint is where a alternative sequence is encountered
# it will be simply inserted into the list of moves and contains a list of 
# alternatives

class SGFChoicepoint:
	def __init__(self):
		self._cp = []
	def add(self, cp):
		self._cp.append(cp)
	def __getitem__(self, i):
		return self._cp[i]
	def __str__(self):
		return "SGFChoicepoint: "  + str(self._cp)
	def __len__(self):
		return len(self._cp)
	

# this class represents a move. A game consists of metadata and double linked SGFMoveNodes
class SGFMoveNode:
	def __init__(self):
		self._next = []
		self._prev = None
		self._color = None
		self._number = None
		self._comment = None
		self._coord = None
	def add(self, var):
		self._next.append(var)
	def __str__(self):
		if len(self._next) == 0:
			return "%s%s" % (self._color, self._coord)
		elif len(self._next) == 1:
			return "%s%s, %s" % (self._color, self._coord, str(self._next[0]))
		else:
			s = ""
			c = 1
			for i in self._next:
				s = s + "\n[%s/%s...%s]" % (c, len(self._next), str(i))
				c = c + 1
			
			return "%s%s !!! %s" % (self._color, self._coord, s)

			
		


# represents a whole game, consisting of metadata, and moves
# moves are presented as a SGFChoicepoint
# SGFChoicepoint[x] may either be a tuple of ([black_moves],[white_moves]) or another SGFChoicepoint
class SGFGame:
	metadata = None
	moves = None
	_bm = None
	_wm = None
	_mainline = None

	# return the whole mainline
	def get_mainline(self):
		if self._mainline == None:
			self._create_mainlines()
		return self._mainline

	# return the main line
	def black_moves_mainline(self):
		if self._bm == None:
			self._create_mainlines()
		return self._bm

	# return the main line
	def white_moves_mainline(self):
		if self._wm == None: 
			self._create_mainlines()
		return self._wm

	def _create_mainlines(self):
		self._bm, self._wm, self._mainline = self._mainlining(self.moves)
		logger.debug("_create_mainlines white: " + str(self._wm))
		logger.debug("_create_mainlines black: " + str(self._bm))
	
	# wanders the SGFMoveNode double linked list along next[0] till it
	# finds an end
	# return:  a tuple
	#	list of white coords forming the white mainline
	#	list of black coords forming the black mainlibne	
	#	list of nodes forming the mainline
	def _mainlining(self, moves):
		wm = []
		bm = []
		mm = []
		node = moves
		logger.debug("_mainlining: %s", node)
		while len(node._next):
			if node._color == 'B':
				bm.append(node._coord)
			elif node._color == 'W':
				wm.append(node._coord)
			else:
				raise "Unknown move color %s" % node._color
			mm.append(node)
			node = node._next[0]
		else:
			# handle the last node too, warning, this could also be the empty list case
			if node:
				if node._color == 'B':
					bm.append(node._coord)
				elif node._color == 'W':
					wm.append(node._coord)
				else:
					raise "Unknown move color %s" % node._color
				mm.append(node)
		return (bm, wm, mm)
	
	def __str__(self):
		return  "SGFGame\n" \
		 	"   metadata: %s\n"  \
		 	"   moves: %s\n " % (str(self.metadata), str(self.moves))

	

class SGFProcessor(simpleparse.dispatchprocessor.DispatchProcessor):
	def __init__(self):
		self._games = []
		pass

	def Collection(self):
		raise "SGFProcessor::Collection"
	# GameTree is the one and only TopLevel Production called by simpleparser, seem like a bug to me
	# but it is can be called more than once
	def GameTree(self, game, buf):
		# all top level elements in children are sequences
		# this is only important for Collections of Games

		# make the buffer avail for all methods
		self._buffer = buf

		pgame = self.process(game)
		# now a "nicer" representation is in self._seq, convert this to a even better suited represenation
		return self._nice_rep(pgame)


	# processes one sequence
	def process(self, gametree):
		(nodename, start, stop, children) = gametree
		if nodename == 'whitespace':
			assert(children == None)
			return 
		elif nodename =='Sequence':
			# we found a sequence, process it and add it to choicepoint
			return self.process_sequence(children)
		elif nodename == 'GameTree':
			cp = SGFChoicepoint()
			for c in children:
				r = self.process(c)
				if r != None:
					cp.add(r)
			return cp
		else:
			raise "unprocessed node" + str(nodename)
				
	
	def process_sequence(self, children):
		seq = []
		for (nodename, start, stop, children2) in children:
			if nodename == 'Node':
				# this is part of the Sequence
				seq.append(self.process_node(children2))
			elif nodename == 'whitespace':
				pass
			elif nodename == 'EmptyNode':
				logger.warning("Warning: forbidden empty node in sgf file - ignoring")
				pass
			else:
				raise "Unknown Node " + str(nodename) + " in process_sequence"
		return seq

	def process_node(self, node):
		nodes = []
		# have to process a list of Properties here
		for (prop, start, stop, children) in node:
			if   prop == 'whitespace':
				pass	#discard
			elif prop == 'Property':
				nodes.append( self.process_prop(children) )
			else:
				raise "Unknown Node " + prop + " in process_node"
		return nodes
	
	def process_prop(self, prop):
		# have to process a list of property values
		pi = ''
		pv = []
		for (name, start, stop, children) in prop:
			if   name == 'whitespace':
				pass
			elif name == 'PropIdent':
				if pi != '':
					raise "Double PropIdent in process_prop"
				pi = self._buffer[start:stop]
			elif name == 'PropValue':
				pv.append(self.process_prop_v(children))
		return (pi,pv)
	
	def process_prop_v(self, val):
		ret = ''
		for (name, start, stop, c) in val:
			if   name == 'whitespace':
				pass
			elif name == 'PropChar':
				ret += self._buffer[start:stop]
				if c != None:
					ret += self.process_prop_v(c)
		return ret 

	# translate the data in game into something even more beautiful
	def _nice_rep(self, game):
		assert(isinstance(game, SGFChoicepoint))

		metadata = SGFGameMeta()
		aw = None	# setup stones, handicap e.g.
		ab = None	# ditto for black
		check1 = True	#FIXME sgf2misc:3.1.9 seems skip GM
		check2 = True	#FIXME noone seems to care about sgf specification, so just read everything
		check3 = True	#FIXME sgf3misc:3.1.9 seems to sktip size

		for meta in game[0][0]:
			# these are the metadata of a game
			name, data = meta
			if   name == 'CA': pass			# char encoding is not important
			elif name == 'GM': 
				if int(data[0]) != 1:
					raise "No Go record"
				check1 = True
			elif name == 'FF': 
				if int(data[0]) != 4:
					#raise "Only version 4 files are supported"	FIXME thinka about that
					pass
				check2 = True
			elif name == 'SZ':
				metadata.size = int(data[0])
				check3 = True
			elif name == 'PW':
				metadata.player_white = data[0]
			elif name == 'PB':
				metadata.player_black = data[0]
			elif name == 'WR':
				metadata.rank_white = data[0]
			elif name == 'BR':
				metadata.rank_black = data[0]
			elif name == 'KM':
				metadata.komi = data[0]
			elif name == 'RE':
				metadata.result = data[0]
			elif name == 'EV':
				metadata.event = data[0]
			elif name == 'DT':
				metadata.date = data[0]
			elif name == 'RO':
				metadata.round = data[0]
			elif name == 'RU':
				metadata.rules = data[0]
			elif name == 'PC':
				metadata.place = data[0]
			elif name == 'TM':
				metadata.timelimit = data[0]
			elif name == 'AW':
				aw = data
			elif name == 'AB':
				ab = data
			else:
				logger.warning('Unknown property %s' % str(name))
		if not (check1 and check2 and check3): raise "Record type not supported"

		# consider setup moves 
		rootnode = None
		oldnode = rootnode
		if aw != None:
			# we have white setup moves, inject them before the game
			for move in aw:
				node = SGFMoveNode()
				node._coord = self._move_to_logical(move, metadata.size)
				node._prev = oldnode
				node._color = 'W'

				if rootnode == None:
					rootnode = node
				else:
					oldnode.add(node)
				oldnode = node
		if ab != None:
			# we have black setup moves, inject them before the game
			for move in ab:
				node = SGFMoveNode()
				node._coord = self._move_to_logical(move, metadata.size)
				node._prev = oldnode
				node._color = 'B'

				if rootnode == None:
					rootnode = node
				else:
					oldnode.add(node)
				oldnode = node
		newrootnode = oldnode		# this "rootnode" is only root for the following game trees
		logger.info("Setup moves found: " + str(rootnode))

		# everything after the metadata are moves, mixed with SGFChoicepoints, may be None!
		m1 = self._nice_rep_game(game[0][1:], metadata.size)

		# append the 
		if rootnode == None:
			rootnode = m1
			newrootnode = rootnode	# this makes adding more games trees easier
		else:
			if m1: newrootnode.add(m1)

		# but there could be additional game trees to consider
		for i in game[1:]:
			if i == None: continue
			m2 = self._nice_rep_game(i, metadata.size)
			# m2 may be None
			if m2: newrootnode.add(m2)


		game = SGFGame()
		game.metadata = metadata
		game.moves = rootnode

		self._games.append( game )
		return self._games

	# build the double linked SGFMoveType structure
	def _nice_rep_game(self, game, size):
		if isinstance(game, SGFChoicepoint):
			node = None
			for i in game:
				if i != None:
					g = self._nice_rep_game(i, size)	# g may be None
					if g: 
						if node == None:
							node = g
						else:
							node.add(g)
			return node
		# we did not have a choicepoint, so we have a sequence of moves
		rootnode = None
		oldnode = None
		for move in game:
			name, data = move[0]
			additional = move[1:]

			node = SGFMoveNode()			# create MoveNode
			node._coord = self._move_to_logical(data[0], size)
			node._additional = additional
			node._prev = oldnode


			if name == 'B':
				node._color = 'B'
			elif name == 'W':
				node._color = 'W'
			else:
				raise "Unknown Property in move " + name

			if rootnode == None:
				rootnode = node
			else:
				oldnode.add(node)
			oldnode = node
		return rootnode
				


#		if isinstance(game, SGFChoicepoint):
#			cp = SGFChoicepoint()
#			for i in game:
#				if i != None:
#					cp.add(self._nice_rep_game(i, size))
#			return cp
#	
#		black_moves = []
#		white_moves = []
#		for move in game:
#			name, data = move[0]
#			additional = move[1:]
#			if   name == 'B':
#				black_moves.append((self._move_to_logical(data[0], size), additional))
#			elif name == 'W':
#				white_moves.append((self._move_to_logical(data[0], size), additional))
#			else:
#				raise "Unknown Property in move " + name
#		return black_moves, white_moves

	def _move_to_logical(self, move, size):
		# converts a sgf file move encoding to internal datastructure
		m = move.upper()
		if len(m) == 0: return (-1, -1)
		x = ord(m[0]) - ord('A')
		y = ord(m[1]) - ord('A')

		return (x,size-y-1)

			
			

# there should be no such thing as an empty Node according to the sgf specifications
# BUT: of course out there are agf files containing empty nodes
# so we read them and throw them away, its a shame
class SGFParser (simpleparse.parser.Parser):
	def __init__(self):
		self._dec = r'''
			Collection	:= GameTree+
			GameTree 	:= whitespace*, "(", Sequence, GameTree* , ")", whitespace*
			Sequence	:= (Node/EmptyNode)+ 
			Node		:= whitespace*, ";", whitespace*, Property+
			EmptyNode	:= whitespace*, ";", whitespace*
			Property	:= PropIdent, whitespace*, PropValue+, whitespace*
			PropIdent	:= [A-Z]+
			PropValue	:= whitespace*, "[", (EscapedPropChar/PropChar)*,"]", whitespace*
			EscapedPropChar := '\\]'
			PropChar	:= -[]]
		
		'''
		simpleparse.parser.Parser.__init__(self, self._dec, 'Collection')
	
	def buildProcessor(self):
		return SGFProcessor()

	# returning (successFlag, resultTree, nextCharacters)
	def parse_file(self, file):
		logger.info("parsing file %s" % file)
		f = open(file, 'r')
		c = f.read()
		return self.parse(c)

	def parse(self, data):
		try:
			logger.info("parsing :" + str(data))
			return simpleparse.parser.Parser.parse(self, data)
		except Exception, e:
			logger.critical("parse " + str(e))
			return False, None, None
	

if __name__ == '__main__':
	import sys
	import os.path

	if os.getenv("GNOMEGO_DEBUG") != None:
		logging.basicConfig(level=logging.DEBUG)
	else:
		logging.basicConfig()

	if len(sys.argv) == 2:
		verbose = False
		files = []
		if os.path.isdir(sys.argv[1]):
			files1 = os.listdir(sys.argv[1])
			for f in files1:
				files.append(sys.argv[1] + f)
		else:
			files =[sys.argv[1]]
			verbose = True
		sgfp = SGFParser()
		succc = 0
		failc = 0
		excpc = 0
		for f in files:
			if verbose:
				res,games,next = sgfp.parse_file(f)
				if not res:
					print "Could not parse file"
					break
				print games[0][0].metadata
				print "Black Moves: " + str(games[0][0].black_moves_mainline())
				print "White Moves: " + str(games[0][0].white_moves_mainline())
				print next
				print "Creating GnomeGoGame"
				g = GnomeGoGame(games)
				print "Stepping through game..."
				while g.step():
					pass
			else:
				try:	
					res,games,next = sgfp.parse_file(f)
					if res:
						print "Success parsing file: " + f
						succc = succc + 1
					else:
						print "Failed parsing file: " + f
						failc = failc + 1
				except:
					print "Exception for file: " + f
					excpc = excpc + 1

		print 
		print "Total files processed: " + str(len(files))
		print "Success: " + str(succc)
		print "Failed: " + str(failc)
		print "Exceptions: " + str(excpc)
	else:
		print "Usage: python sgfparser.py /path/to/sgffiles"
		print "       trys to parse each file in given directory, prints stats at the end"
		print "Usage: python sgfparser.py /path/to/file.sgf"
		print "       trys to parse that single given file"
		print "       output is more verbose"
		sys.exit(-1)
